@@ -3,17 +3,23 @@
 #include "ev-page-cache.h"
 #include "ev-selection.h"
 #include "ev-document-images.h"
+#include "ev-document-forms.h"
 #include "ev-image.h"
+#include "ev-form-field.h"
 
 typedef struct _CacheJobInfo
 {
 	EvJob *job;
 	EvRenderContext *rc;
 
+	/* Region of the page that needs to be drawn */
+	GdkRegion *region; 
+
 	/* Data we get from rendering */
-	GdkPixbuf *pixbuf;
+	cairo_surface_t *surface;
 	GList *link_mapping;
 	GList *image_mapping;
+	GList *form_field_mapping;
 	GdkRegion *text_mapping;
 	
 	/* Selection data. 
@@ -23,7 +29,7 @@ typedef struct _CacheJobInfo
 	EvRectangle target_points;
 	gboolean    points_set;
 	
-	GdkPixbuf *selection;
+	cairo_surface_t *selection;
 	GdkRegion *selection_region;
 } CacheJobInfo;
 
@@ -74,10 +80,10 @@ static CacheJobInfo *find_job_cache             (EvPixbufCache      *pixbuf_cach
 static void          copy_job_to_job_info       (EvJobRender        *job_render,
 						 CacheJobInfo       *job_info,
 						 EvPixbufCache      *pixbuf_cache);
-static gboolean      new_selection_pixbuf_needed(EvPixbufCache      *pixbuf_cache,
-						 CacheJobInfo       *job_info,
-						 gint                page,
-						 gfloat              scale);
+static gboolean      new_selection_surface_needed(EvPixbufCache      *pixbuf_cache,
+						  CacheJobInfo       *job_info,
+						  gint                page,
+						  gfloat              scale);
 
 
 /* These are used for iterating through the prev and next arrays */
@@ -112,13 +118,15 @@ ev_pixbuf_cache_class_init (EvPixbufCacheClass *class)
 	object_class->finalize = ev_pixbuf_cache_finalize;
 	object_class->dispose = ev_pixbuf_cache_dispose;
 
-	signals[JOB_FINISHED] = g_signal_new ("job-finished",
-					    G_OBJECT_CLASS_TYPE (object_class),
-					    G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
-					    G_STRUCT_OFFSET (EvPixbufCacheClass, job_finished),
-					    NULL, NULL,
-					    g_cclosure_marshal_VOID__VOID,
-					    G_TYPE_NONE, 0);
+	signals[JOB_FINISHED] =
+		g_signal_new ("job-finished",
+			      G_OBJECT_CLASS_TYPE (object_class),
+			      G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
+			      G_STRUCT_OFFSET (EvPixbufCacheClass, job_finished),
+			      NULL, NULL,
+			      g_cclosure_marshal_VOID__POINTER,
+			      G_TYPE_NONE, 1,
+			      G_TYPE_POINTER);
 }
 
 static void
@@ -149,9 +157,13 @@ dispose_cache_job_info (CacheJobInfo *job_info,
 		g_object_unref (G_OBJECT (job_info->job));
 		job_info->job = NULL;
 	}
-	if (job_info->pixbuf) {
-		g_object_unref (G_OBJECT (job_info->pixbuf));
-		job_info->pixbuf = NULL;
+	if (job_info->surface) {
+		cairo_surface_destroy (job_info->surface);
+		job_info->surface = NULL;
+	}
+	if (job_info->region) {
+		gdk_region_destroy (job_info->region);
+		job_info->region = NULL;
 	}
 	if (job_info->link_mapping) {
 		ev_link_mapping_free (job_info->link_mapping);
@@ -161,12 +173,16 @@ dispose_cache_job_info (CacheJobInfo *job_info,
 		ev_image_mapping_free (job_info->image_mapping);
 		job_info->image_mapping = NULL;
 	}
+	if (job_info->form_field_mapping) {
+		ev_form_field_mapping_free (job_info->form_field_mapping);
+		job_info->form_field_mapping = NULL;
+	}
 	if (job_info->text_mapping) {
 		gdk_region_destroy (job_info->text_mapping);
 		job_info->text_mapping = NULL;
 	}
 	if (job_info->selection) {
-		g_object_unref (G_OBJECT (job_info->selection));
+		cairo_surface_destroy (job_info->selection);
 		job_info->selection = NULL;
 	}
 	if (job_info->selection_region) {
@@ -233,8 +249,7 @@ job_finished_cb (EvJob         *job,
 	job_info = find_job_cache (pixbuf_cache, job_render->rc->page);
 
 	copy_job_to_job_info (job_render, job_info, pixbuf_cache);
-
-	g_signal_emit (pixbuf_cache, signals[JOB_FINISHED], 0);
+	g_signal_emit (pixbuf_cache, signals[JOB_FINISHED], 0, job_info->region);
 }
 
 /* This checks a job to see if the job would generate the right sized pixbuf
@@ -321,17 +336,17 @@ move_one_job (CacheJobInfo  *job_info,
 
 	*target_page = *job_info;
 	job_info->job = NULL;
-	job_info->pixbuf = NULL;
+	job_info->region = NULL;
+	job_info->surface = NULL;
 	job_info->link_mapping = NULL;
 	job_info->image_mapping = NULL;
+	job_info->form_field_mapping = NULL;
 
 	if (new_priority != priority && target_page->job) {
 		ev_job_queue_update_job (target_page->job, new_priority);
 	}
 }
 
-
-
 static void
 ev_pixbuf_cache_update_range (EvPixbufCache *pixbuf_cache,
 			      gint           start_page,
@@ -411,10 +426,10 @@ copy_job_to_job_info (EvJobRender   *job_render,
 
 	job_info->points_set = FALSE;
 
-	if (job_info->pixbuf) {
-		g_object_unref (G_OBJECT (job_info->pixbuf));
+	if (job_info->surface) {
+		cairo_surface_destroy (job_info->surface);
 	}
-	job_info->pixbuf = g_object_ref (job_render->pixbuf);
+	job_info->surface = cairo_surface_reference (job_render->surface);
 
 	if (job_info->rc) {
 		g_object_unref (G_OBJECT (job_info->rc));
@@ -433,6 +448,12 @@ copy_job_to_job_info (EvJobRender   *job_render,
 		job_info->image_mapping = job_render->image_mapping;
 	}
 
+	if (job_render->include_forms) {
+		if (job_info->form_field_mapping)
+			ev_form_field_mapping_free (job_info->form_field_mapping);
+		job_info->form_field_mapping = job_render->form_field_mapping;
+	}
+
 	if (job_render->include_text) {
 		if (job_info->text_mapping)
 			gdk_region_destroy (job_info->text_mapping);
@@ -441,7 +462,7 @@ copy_job_to_job_info (EvJobRender   *job_render,
 
 	if (job_render->include_selection) {
 		if (job_info->selection) {
-			g_object_unref (G_OBJECT (job_info->selection));
+			cairo_surface_destroy (job_info->selection);
 			job_info->selection = NULL;
 		}
 		if (job_info->selection_region) {
@@ -451,8 +472,9 @@ copy_job_to_job_info (EvJobRender   *job_render,
 		
 		job_info->selection_points = job_render->selection_points;
 		job_info->selection_region = gdk_region_copy (job_render->selection_region);
-		job_info->selection = g_object_ref (job_render->selection);
+		job_info->selection = cairo_surface_reference (job_render->selection);
 		g_assert (job_info->selection_points.x1 >= 0);
+		job_info->points_set = TRUE;
 	}
 
 	if (job_info->job) {
@@ -532,33 +554,24 @@ get_selection_colors (GtkWidget *widget, GdkColor **text, GdkColor **base)
 }
 
 static void
-add_job_if_needed (EvPixbufCache *pixbuf_cache,
-		   CacheJobInfo  *job_info,
-		   EvPageCache   *page_cache,
-		   gint           page,
-		   gint           rotation,
-		   gfloat         scale,
-		   EvJobPriority  priority)
+add_job (EvPixbufCache *pixbuf_cache,
+	 CacheJobInfo  *job_info,
+	 EvPageCache   *page_cache,
+	 GdkRegion     *region,
+	 gint           width,
+	 gint           height,
+	 gint           page,
+	 gint           rotation,
+	 gfloat         scale,
+	 EvJobPriority  priority)
 {
 	gboolean include_links = FALSE;
 	gboolean include_text = FALSE;
 	gboolean include_selection = FALSE;
-	gboolean include_images = FALSE;
-	int width, height;
+	gboolean include_images = TRUE;
+	gboolean include_forms = FALSE;
 	GdkColor *text, *base;
-
-	if (job_info->job)
-		return;
-
-	ev_page_cache_get_size (page_cache, page, rotation,
-				scale, &width, &height);
-
-	if (job_info->pixbuf &&
-	    gdk_pixbuf_get_width (job_info->pixbuf) == width &&
-	    gdk_pixbuf_get_height (job_info->pixbuf) == height)
-		return;
-
-	/* make a new job now */
+	
 	if (job_info->rc == NULL) {
 		job_info->rc = ev_render_context_new (rotation, page, scale);
 	} else {
@@ -567,14 +580,20 @@ add_job_if_needed (EvPixbufCache *pixbuf_cache,
 		ev_render_context_set_scale (job_info->rc, scale);
 	}
 
+	if (job_info->region)
+		gdk_region_destroy (job_info->region);
+	job_info->region = region ? gdk_region_copy (region) : NULL;
+
 	/* Figure out what else we need for this job */
 	if (job_info->link_mapping == NULL)
 		include_links = TRUE;
 	if (job_info->image_mapping == NULL)
 		include_images = TRUE;
+	if (job_info->form_field_mapping == NULL)
+		include_forms = TRUE;
 	if (job_info->text_mapping == NULL)
 		include_text = TRUE;
-	if (new_selection_pixbuf_needed (pixbuf_cache, job_info, page, scale)) {
+	if (new_selection_surface_needed (pixbuf_cache, job_info, page, scale)) {
 		include_selection = TRUE;
 	}
 
@@ -587,14 +606,43 @@ add_job_if_needed (EvPixbufCache *pixbuf_cache,
 					   width, height,
 					   &(job_info->target_points),
 					   text, base,
+					   include_forms,
 					   include_links,
 					   include_images,
 					   include_text,
 					   include_selection);
 	ev_job_queue_add_job (job_info->job, priority);
-	g_signal_connect (job_info->job, "finished", G_CALLBACK (job_finished_cb), pixbuf_cache);
+	g_signal_connect (job_info->job, "finished",
+			  G_CALLBACK (job_finished_cb),
+			  pixbuf_cache);
 }
 
+static void
+add_job_if_needed (EvPixbufCache *pixbuf_cache,
+		   CacheJobInfo  *job_info,
+		   EvPageCache   *page_cache,
+		   gint           page,
+		   gint           rotation,
+		   gfloat         scale,
+		   EvJobPriority  priority)
+{
+	gint width, height;
+
+	if (job_info->job)
+		return;
+
+	ev_page_cache_get_size (page_cache, page, rotation,
+				scale, &width, &height);
+
+	if (job_info->surface &&
+	    cairo_image_surface_get_width (job_info->surface) == width &&
+	    cairo_image_surface_get_height (job_info->surface) == height)
+		return;
+
+	add_job (pixbuf_cache, job_info, page_cache, NULL,
+		 width, height, page, rotation, scale,
+		 priority);
+}
 
 static void
 ev_pixbuf_cache_add_jobs_if_needed (EvPixbufCache *pixbuf_cache,
@@ -671,9 +719,9 @@ ev_pixbuf_cache_set_page_range (EvPixbufCache  *pixbuf_cache,
 	ev_pixbuf_cache_add_jobs_if_needed (pixbuf_cache, rotation, scale);
 }
 
-GdkPixbuf *
-ev_pixbuf_cache_get_pixbuf (EvPixbufCache *pixbuf_cache,
-			    gint           page)
+cairo_surface_t *
+ev_pixbuf_cache_get_surface (EvPixbufCache *pixbuf_cache,
+			     gint           page)
 {
 	CacheJobInfo *job_info;
 
@@ -685,9 +733,10 @@ ev_pixbuf_cache_get_pixbuf (EvPixbufCache *pixbuf_cache,
 	if (job_info->job &&
 	    EV_JOB (job_info->job)->finished) {
 		copy_job_to_job_info (EV_JOB_RENDER (job_info->job), job_info, pixbuf_cache);
+		g_signal_emit (pixbuf_cache, signals[JOB_FINISHED], 0, job_info->region);
 	}
 
-	return job_info->pixbuf;
+	return job_info->surface;
 }
 
 GList *
@@ -704,6 +753,7 @@ ev_pixbuf_cache_get_link_mapping (EvPixbufCache *pixbuf_cache,
 	if (job_info->job &&
 	    EV_JOB (job_info->job)->finished) {
 		copy_job_to_job_info (EV_JOB_RENDER (job_info->job), job_info, pixbuf_cache);
+		g_signal_emit (pixbuf_cache, signals[JOB_FINISHED], 0, job_info->region);
 	}
 
 	return job_info->link_mapping;
@@ -726,32 +776,62 @@ ev_pixbuf_cache_get_image_mapping (EvPixbufCache *pixbuf_cache,
 	if (job_info->job &&
 	    EV_JOB (job_info->job)->finished) {
 		copy_job_to_job_info (EV_JOB_RENDER (job_info->job), job_info, pixbuf_cache);
+		g_signal_emit (pixbuf_cache, signals[JOB_FINISHED], 0, job_info->region);
 	}
 
 	return job_info->image_mapping;
 }
 
+GList *
+ev_pixbuf_cache_get_form_field_mapping (EvPixbufCache *pixbuf_cache,
+				  	gint 	       page)
+{
+	CacheJobInfo *job_info;
+
+	if (!EV_IS_DOCUMENT_FORMS (pixbuf_cache->document))
+		return NULL;
+	
+	job_info = find_job_cache (pixbuf_cache, page);
+	if (job_info == NULL)
+		return NULL;
+
+	/* We don't need to wait for the idle to handle the callback */
+	if (job_info->job &&
+	   EV_JOB (job_info->job)->finished) {
+		copy_job_to_job_info (EV_JOB_RENDER(job_info->job), job_info, pixbuf_cache);
+		g_signal_emit (pixbuf_cache, signals[JOB_FINISHED], 0, job_info->region);
+	}
+	
+	return job_info->form_field_mapping;
+}
+
 static gboolean
-new_selection_pixbuf_needed (EvPixbufCache *pixbuf_cache,
+new_selection_surface_needed (EvPixbufCache *pixbuf_cache,
 			     CacheJobInfo  *job_info,
 			     gint           page,
 			     gfloat         scale)
 {
 	EvPageCache *page_cache;
-	gint width, height;
 
 	if (job_info->selection) {
-		page_cache = ev_page_cache_get (pixbuf_cache->document);
-		ev_page_cache_get_size (page_cache, page, job_info->rc->rotation,
-					scale, &width, &height);
+		gint width, height;
+		gint selection_width, selection_height;
 		
-		if (width != gdk_pixbuf_get_width (job_info->selection) ||
-		    height != gdk_pixbuf_get_height (job_info->selection))
+		page_cache = ev_page_cache_get (pixbuf_cache->document);
+		ev_page_cache_get_size (page_cache, page,
+					job_info->rc->rotation,
+					scale, &width, &height);
+
+		selection_width = cairo_image_surface_get_width (job_info->selection);
+		selection_height = cairo_image_surface_get_height (job_info->selection);
+		
+		if (width != selection_width || height != selection_height)
 			return TRUE;
 	} else {
 		if (job_info->points_set)
 			return TRUE;
 	}
+	
 	return FALSE;
 }
 
@@ -761,9 +841,9 @@ clear_selection_if_needed (EvPixbufCache *pixbuf_cache,
 			   gint           page,
 			   gfloat         scale)
 {
-	if (new_selection_pixbuf_needed (pixbuf_cache, job_info, page, scale)) {
+	if (new_selection_surface_needed (pixbuf_cache, job_info, page, scale)) {
 		if (job_info->selection)
-			g_object_unref (job_info->selection);
+			cairo_surface_destroy (job_info->selection);
 		job_info->selection = NULL;
 		job_info->selection_points.x1 = -1;
 	}
@@ -783,6 +863,7 @@ ev_pixbuf_cache_get_text_mapping (EvPixbufCache *pixbuf_cache,
 	if (job_info->job &&
 	    EV_JOB (job_info->job)->finished) {
 		copy_job_to_job_info (EV_JOB_RENDER (job_info->job), job_info, pixbuf_cache);
+		g_signal_emit (pixbuf_cache, signals[JOB_FINISHED], 0, job_info->region);
 	}
 	
 	return job_info->text_mapping;
@@ -817,13 +898,13 @@ ev_pixbuf_cache_style_changed (EvPixbufCache *pixbuf_cache)
 
 		job_info = pixbuf_cache->prev_job + i;
 		if (job_info->selection) {
-			g_object_unref (G_OBJECT (job_info->selection));
+			cairo_surface_destroy (job_info->selection);
 			job_info->selection = NULL;
 		}
 
 		job_info = pixbuf_cache->next_job + i;
 		if (job_info->selection) {
-			g_object_unref (G_OBJECT (job_info->selection));
+			cairo_surface_destroy (job_info->selection);
 			job_info->selection = NULL;
 		}
 	}
@@ -833,17 +914,17 @@ ev_pixbuf_cache_style_changed (EvPixbufCache *pixbuf_cache)
 
 		job_info = pixbuf_cache->job_list + i;
 		if (job_info->selection) {
-			g_object_unref (G_OBJECT (job_info->selection));
+			cairo_surface_destroy (job_info->selection);
 			job_info->selection = NULL;
 		}
 	}
 }
 
-GdkPixbuf *
-ev_pixbuf_cache_get_selection_pixbuf (EvPixbufCache  *pixbuf_cache,
-				      gint            page,
-				      gfloat          scale,
-				      GdkRegion     **region)
+cairo_surface_t *
+ev_pixbuf_cache_get_selection_surface (EvPixbufCache  *pixbuf_cache,
+				       gint            page,
+				       gfloat          scale,
+				       GdkRegion     **region)
 {
 	CacheJobInfo *job_info;
 
@@ -931,7 +1012,7 @@ clear_job_selection (CacheJobInfo *job_info)
 	job_info->selection_points.x1 = -1;
 
 	if (job_info->selection) {
-		g_object_unref (job_info->selection);
+		cairo_surface_destroy (job_info->selection);
 		job_info->selection = NULL;
 	}
 }
@@ -1092,3 +1173,28 @@ ev_pixbuf_cache_get_selection_list (EvPixbufCache *pixbuf_cache)
 	return retval;
 }
 
+void           
+ev_pixbuf_cache_reload_page (EvPixbufCache *pixbuf_cache,
+			     GdkRegion     *region,
+			     gint           page,
+			     gint           rotation,
+			     gdouble        scale)
+{
+	CacheJobInfo *job_info;
+        EvPageCache *page_cache;
+        gint width, height;
+
+	job_info = find_job_cache (pixbuf_cache, page);
+	if (job_info == NULL)
+		return;
+	
+        page_cache = ev_page_cache_get (pixbuf_cache->document);
+	ev_page_cache_get_size (page_cache, page, rotation, scale,
+				&width, &height);
+
+        add_job (pixbuf_cache, job_info, page_cache, region,
+		 width, height, page, rotation, scale,
+		 EV_JOB_PRIORITY_HIGH);
+}
+
+
